// Copyright (c) 2006, TopCoder, Inc. All rights reserved.
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using TopCoder.Uml.Xmi.MetaData;

namespace TopCoder.Uml.Xmi
{
    /// <summary>
    /// Provides helper methods for the xmi2cs application.
    /// </summary>
    /// <author>Dan Manges (Seiz3r)</author>
    /// <copyright>Copyright (c) 2006, TopCoder, Inc. All rights reserved.</copyright>
    internal class Helper
    {
        // The maximum line length to use in source files.
        private const int LINE_LENGTH = 120;

        /// <summary>
        /// Read an attribute from a node.
        /// </summary>
        /// <param name="node">The node to read the attribute from.</param>
        /// <param name="attribute">The attribute to read.</param>
        /// <returns>The attribute value or null.</returns>
        public static string ReadAttribute(XmlNode node, string attribute)
        {
            return ReadAttribute(node, attribute, null);
        }

        /// <summary>
        /// Read an attribute from a node, specifiying a return value if the attribute does not exist.
        /// </summary>
        /// <param name="node">The node to read the attribute from.</param>
        /// <param name="attribute">The attribute to read.</param>
        /// <param name="valueIfNull">The value to return if the attribute does not exist.</param>
        /// <returns>The attribute value or the valueIfNull parameter.</returns>
        public static string ReadAttribute(XmlNode node, string attribute, string valueIfNull)
        {
            XmlAttribute xmlAttribute = node.Attributes[attribute];
            if (xmlAttribute == null)
            {
                return valueIfNull;
            }
            else
            {
                return xmlAttribute.Value;
            }
        }

        /// <summary>
        /// Joins a namespace with a class name, returning only the class name if the namespace is empty.
        /// </summary>
        /// <remarks>
        /// Prevents a simple concatenation from resulting in ".ClassName" when the namespace is empty.
        /// </remarks>
        /// <param name="namespace">The namespace to join.</param>
        /// <param name="className">The class name to join.</param>
        /// <returns>The namespace and class name properly concatenated.</returns>
        public static string JoinNamespace(string @namespace, string className)
        {
            if (@namespace == "")
            {
                return className;
            }
            return @namespace + "." + className;
        }

        /// <summary>
        /// The file name to use for a class.
        /// </summary>
        /// <remarks>
        /// Although the result is simple, this will make it easier to change destination later, such as if the source
        /// file were placed in folders by namespace.
        /// </remarks>
        /// <param name="directory">The target directory.</param>
        /// <param name="namespace">The namespace for the class.</param>
        /// <param name="className">The class name.</param>
        /// <param name="extension">The file extension.</param>
        /// <returns>The filename based on the parameters.</returns>
        public static string FileForClass(string directory, string @namespace, string className, string extension)
        {
            @namespace = GetValidFilename(@namespace);
            className = GetValidFilename(className);
            string path = directory + @"\" + @namespace.Replace('.', '\\');
            if (!Directory.Exists(path))
            {
                Directory.CreateDirectory(path);
            }
            return path + @"\" + className + "." + extension;
        }


        /// <summary>
        /// Sanitize file names.
        /// </summary>
        /// <param name="input">The string to replace invalid file name characters.</param>
        /// <returns>A safe filename.</returns>
        public static string GetValidFilename(string input)
        {
            string invalidChars = "/\\:*?\"<>|";
            foreach (char c in invalidChars)
            {
                input = input.Replace(c, '_');
            }
            return input;
        }

        /// <summary>
        /// Convert a XML node name string to a UML type to determine node references.
        /// </summary>
        /// <param name="input">The node name string.</param>
        /// <returns>The UmlType.</returns>
        public static UmlType StringToUmlType(string input)
        {
            switch (input)
            {
                case "UML:DataType":
                case "DataType":
                    return UmlType.DataType;
                case "UML:Class":
                case "Class":
                    return UmlType.Class;
                case "UML:Interface":
                case "Interface":
                    return UmlType.Interface;
                default:
                    return UmlType.Other;
            }
        }

        /// <summary>
        /// Convert a documentation type string into a specific DocumentationType.
        /// </summary>
        /// <param name="input">The documentation type string.</param>
        /// <returns>A specific DocumentationType.</returns>
        public static DocumentationType StringToDocumentationType(string input)
        {
            switch (input)
            {
                case "documentation":
                    return DocumentationType.Summary;
                case "documentation#returns":
                    return DocumentationType.Returns;
                case "documentation#throws":
                    return DocumentationType.Throws;
                case "documentation#param value":
                    return DocumentationType.ParamValue;
                default:
                    return DocumentationType.Other;
            }
        }

        /// <summary>
        /// Get source code file formatted documentation from an XmiElement.
        /// </summary>
        /// <param name="xmiElement">The element to read documentation from.</param>
        /// <param name="leadingSpaces">The number of spaces to indent the documentation.</param>
        /// <returns>The formatted documentation.</returns>
        public static string GetDocumentation(XmiElement xmiElement, int leadingSpaces)
        {
            // Store documentation line by line until the end.
            List<string> lines = new List<string>();
            // Loop through each UmlTaggedValue, which represents a documentation node.
            foreach (UmlTaggedValue umlTaggedValue in xmiElement.UmlTaggedValues)
            {
                // Variables to store the type of documentation tag.
                string startTag;
                string endTag;
                switch (umlTaggedValue.DocumentationType)
                {
                    // If the type is summary, make a summary node unless the element is a parameter.
                    case DocumentationType.Summary:
                        if (xmiElement is UmlParameter)
                        {
                            // If it is a return parameter, use a return documentation tag.
                            if (((UmlParameter)xmiElement).Kind == "return")
                            {
                                startTag = "<returns>";
                                endTag = "</returns>";
                            }
                            else
                            {
                                // If a non-return parameter, use a param tag.
                                startTag = String.Format("<param name=\"{0}\">", ((UmlParameter)xmiElement).Name);
                                endTag = "</param>";
                            }
                        }
                        else
                        {
                            // If not a parameter, use a summary tag.
                            startTag = "<summary>";
                            endTag = "</summary>";
                        }
                        break;
                    case DocumentationType.Returns:
                        startTag = "<returns>";
                        endTag = "</returns>";
                        break;
                    case DocumentationType.Throws:
                        // Try to find the Exception class name.
                        Match classNameMatch = Regex.Match(umlTaggedValue.Value, "([A-Za-z]+Exception)");
                        if (classNameMatch.Success)
                        {
                            // If an exception class name is found, put it in the documentation.
                            startTag = String.Format("<exception cref=\"{0}\">", classNameMatch.Groups[1].Value);
                        }
                        else
                        {
                            // Otherwise do not output a cref tag.
                            startTag = "<exception>";
                        }
                        endTag = "</exception>";
                        break;
                    case DocumentationType.ParamValue:
                        if (xmiElement is UmlParameter)
                        {
                            if (((UmlParameter)xmiElement).Kind == "return")
                            {
                                startTag = "<returns>";
                                endTag = "</returns>";
                            }
                            else
                            {
                                startTag = String.Format("<param name=\"{0}\">", ((UmlParameter)xmiElement).Name);
                                endTag = "</param>";
                            }
                        }
                        else
                        {
                            startTag = "<value>";
                            endTag = "</value>";
                        }
                        break;
                    // Place unknown documentation types in a "remarks" tag.
                    case DocumentationType.Other:
                    default:
                        startTag = "<remarks>";
                        endTag = "</remarks>";
                        break;
                }
                if (leadingSpaces + startTag.Length + umlTaggedValue.Value.Length + endTag.Length < LINE_LENGTH)
                {
                    // If everything fits on one line, format it that way.
                    lines.AddRange(Helper.FormatText(startTag + umlTaggedValue.Value + endTag, leadingSpaces, LINE_LENGTH));
                }
                else
                {
                    // Place the documentation text and start and end nodes on seprate lines.
                    lines.AddRange(Helper.FormatText(startTag, leadingSpaces, LINE_LENGTH));
                    lines.AddRange(Helper.FormatText(umlTaggedValue.Value, leadingSpaces, LINE_LENGTH));
                    lines.AddRange(Helper.FormatText(endTag, leadingSpaces, LINE_LENGTH));
                }
            }
            // Build the documentation lines into a single string.
            StringBuilder builder = new StringBuilder();
            foreach (string line in lines)
            {
                builder.AppendLine(line);
            }
            return builder.ToString();
        }

        /// <summary>
        /// Translate OwnerScope from XMI to a csharp modifier.
        /// </summary>
        /// <param name="input">The ownerScope input.</param>
        /// <returns>The csharp modifier.</returns>
        public static string OwnerScopeToCSharp(string input)
        {
            switch (input)
            {
                // Classifier is static.
                case "classifier":
                    return "static ";
                // Instance does not need displayed.
                case "instance":
                    return "";
                default:
                    return input;
            }
        }

        /// <summary>
        /// Format a string into documentation format.
        /// </summary>
        /// <param name="text">The text to format.</param>
        /// <param name="leadingSpaces">The number of spaces to indent, counting the "/// " characters.</param>
        /// <param name="maxLineLength">The maximum line length for the documentation.</param>
        /// <returns>Formatted documentation.</returns>
        public static string[] FormatText(string text, int leadingSpaces, int maxLineLength)
        {
            // Build a character array of line prefix characters.
            char[] linePrefixCharacters = new char[leadingSpaces];
            for (int i = 0; i < leadingSpaces; i++)
            {
                if (i >= (leadingSpaces - 4) && i <= leadingSpaces - 2)
                {
                    linePrefixCharacters[i] = '/';
                }
                else
                {
                    linePrefixCharacters[i] = ' ';
                }
            }
            // The prefix for each documentation line.
            string linePrefix = new String(linePrefixCharacters);
            // Strip newlines in documentation.
            text = text.Replace("\r", " ").Replace("\n", " ");
            // The number of characters to be used after the prefix.
            int lineLength = maxLineLength - leadingSpaces;
            List<string> lines = new List<string>();
            // The position in the string.
            int position = 0;
            while (position < text.Length)
            {
                // If the remaining text fits on one line, add it.
                if (text.Length - position < lineLength)
                {
                    lines.Add(linePrefix + text.Substring(position, text.Length - position));
                    break;
                }
                // Search for the last space in the limit of the next line.
                int searchLength = Math.Min(lineLength, text.Length - position - 1);
                int bestSpaceIndex = text.LastIndexOf(' ', position + searchLength, searchLength);
                if (bestSpaceIndex == -1)
                {
                    // If a space cannot be found, break the text.
                    lines.Add(linePrefix + text.Substring(position, searchLength + 1).Trim());
                    position += searchLength + 1;
                }
                else
                {
                    // Split the text on the best possible space.
                    lines.Add(linePrefix + text.Substring(position, bestSpaceIndex - position).Trim());
                    position = bestSpaceIndex + 1;
                }
            }
            return lines.ToArray();
        }

        /// <summary>
        /// Extract the namespace from a fully qualified class name.
        /// </summary>
        /// <param name="className">The class name.</param>
        /// <returns>The namespace for the class, or the class name if the namespace cannot be determined.</returns>
        public static string ExtractNamespace(string className)
        {
            int length = className.LastIndexOf('.');
            if (length == -1)
            {
                return className;
            }
            else
            {
                string ns = className.Substring(0, length);
                if (ns.Contains("TopCoder"))
                {
                    return ns;
                }
                if (!ns.Contains("System"))
                {
                    return "System." + ns;
                }
                return ns;
            }
        }
    }
}